@@ -1,22 +1,59 @@
 use crate::resource;
 
-pub fn is_power_of_two(val: u32) -> bool {
+pub fn is_power_of_two_u16(val: u16) -> bool {
     val != 0 && (val & (val - 1)) == 0
 }
 
-pub fn is_valid_copy_src_texture_format(format: wgt::TextureFormat) -> bool {
+pub fn is_power_of_two_u32(val: u32) -> bool {
+    val != 0 && (val & (val - 1)) == 0
+}
+
+pub fn is_valid_copy_src_texture_format(
+    format: wgt::TextureFormat,
+    aspect: wgt::TextureAspect,
+) -> bool {
+    use wgt::TextureAspect as Ta;
     use wgt::TextureFormat as Tf;
-    match format {
-        Tf::Depth24Plus | Tf::Depth24PlusStencil8 => false,
+    match (format, aspect) {
+        (Tf::Depth24Plus, _) | (Tf::Depth24PlusStencil8, Ta::DepthOnly) => false,
         _ => true,
     }
 }
 
-pub fn is_valid_copy_dst_texture_format(format: wgt::TextureFormat) -> bool {
+pub fn is_valid_copy_dst_texture_format(
+    format: wgt::TextureFormat,
+    aspect: wgt::TextureAspect,
+) -> bool {
+    use wgt::TextureAspect as Ta;
+    use wgt::TextureFormat as Tf;
+    match (format, aspect) {
+        (Tf::Depth24Plus | Tf::Depth32Float, _)
+        | (Tf::Depth24PlusStencil8 | Tf::Depth32FloatStencil8, Ta::DepthOnly) => false,
+        _ => true,
+    }
+}
+
+#[cfg_attr(
+    any(not(target_arch = "wasm32"), target_os = "emscripten"),
+    allow(unused)
+)]
+pub fn is_valid_external_image_copy_dst_texture_format(format: wgt::TextureFormat) -> bool {
     use wgt::TextureFormat as Tf;
     match format {
-        Tf::Depth32Float | Tf::Depth24Plus | Tf::Depth24PlusStencil8 => false,
-        _ => true,
+        Tf::R8Unorm
+        | Tf::R16Float
+        | Tf::R32Float
+        | Tf::Rg8Unorm
+        | Tf::Rg16Float
+        | Tf::Rg32Float
+        | Tf::Rgba8Unorm
+        | Tf::Rgba8UnormSrgb
+        | Tf::Bgra8Unorm
+        | Tf::Bgra8UnormSrgb
+        | Tf::Rgb10a2Unorm
+        | Tf::Rgba16Float
+        | Tf::Rgba32Float => true,
+        _ => false,
     }
 }
 
@@ -51,13 +88,17 @@ pub fn map_buffer_usage(usage: wgt::BufferUsages) -> hal::BufferUses {
         usage.contains(wgt::BufferUsages::UNIFORM),
     );
     u.set(
-        hal::BufferUses::STORAGE_READ | hal::BufferUses::STORAGE_WRITE,
+        hal::BufferUses::STORAGE_READ | hal::BufferUses::STORAGE_READ_WRITE,
         usage.contains(wgt::BufferUsages::STORAGE),
     );
     u.set(
         hal::BufferUses::INDIRECT,
         usage.contains(wgt::BufferUsages::INDIRECT),
     );
+    u.set(
+        hal::BufferUses::QUERY_RESOLVE,
+        usage.contains(wgt::BufferUsages::QUERY_RESOLVE),
+    );
     u
 }
 
@@ -70,12 +111,16 @@ pub fn map_texture_usage(
         hal::TextureUses::COPY_SRC,
         usage.contains(wgt::TextureUsages::COPY_SRC),
     );
+    u.set(
+        hal::TextureUses::COPY_DST,
+        usage.contains(wgt::TextureUsages::COPY_DST),
+    );
     u.set(
         hal::TextureUses::RESOURCE,
         usage.contains(wgt::TextureUsages::TEXTURE_BINDING),
     );
     u.set(
-        hal::TextureUses::STORAGE_READ | hal::TextureUses::STORAGE_WRITE,
+        hal::TextureUses::STORAGE_READ | hal::TextureUses::STORAGE_READ_WRITE,
         usage.contains(wgt::TextureUsages::STORAGE_BINDING),
     );
     let is_color = aspect.contains(hal::FormatAspects::COLOR);
@@ -90,6 +135,31 @@ pub fn map_texture_usage(
     u
 }
 
+pub fn map_texture_usage_from_hal(uses: hal::TextureUses) -> wgt::TextureUsages {
+    let mut u = wgt::TextureUsages::empty();
+    u.set(
+        wgt::TextureUsages::COPY_SRC,
+        uses.contains(hal::TextureUses::COPY_SRC),
+    );
+    u.set(
+        wgt::TextureUsages::COPY_DST,
+        uses.contains(hal::TextureUses::COPY_DST),
+    );
+    u.set(
+        wgt::TextureUsages::TEXTURE_BINDING,
+        uses.contains(hal::TextureUses::RESOURCE),
+    );
+    u.set(
+        wgt::TextureUsages::STORAGE_BINDING,
+        uses.contains(hal::TextureUses::STORAGE_READ | hal::TextureUses::STORAGE_READ_WRITE),
+    );
+    u.set(
+        wgt::TextureUsages::RENDER_ATTACHMENT,
+        uses.contains(hal::TextureUses::COLOR_TARGET),
+    );
+    u
+}
+
 pub fn check_texture_dimension_size(
     dimension: wgt::TextureDimension,
     wgt::Extent3d {
@@ -104,14 +174,7 @@ pub fn check_texture_dimension_size(
     use wgt::TextureDimension::*;
 
     let (extent_limits, sample_limit) = match dimension {
-        D1 => (
-            [
-                limits.max_texture_dimension_1d,
-                1,
-                limits.max_texture_array_layers,
-            ],
-            1,
-        ),
+        D1 => ([limits.max_texture_dimension_1d, 1, 1], 1),
         D2 => (
             [
                 limits.max_texture_dimension_2d,
@@ -142,7 +205,7 @@ pub fn check_texture_dimension_size(
             return Err(Tde::LimitExceeded { dim, given, limit });
         }
     }
-    if sample_size == 0 || sample_size > sample_limit || !is_power_of_two(sample_size) {
+    if sample_size == 0 || sample_size > sample_limit || !is_power_of_two_u32(sample_size) {
         return Err(Tde::InvalidSampleCount(sample_size));
     }
 
