@@ -15,13 +15,15 @@ use std::borrow::Borrow;
 use crate::device::trace::Action;
 use crate::{
     conv,
-    device::DeviceError,
-    hub::{Global, GlobalIdentityHandlerFactory, HalApi, Input, Token},
+    device::{DeviceError, MissingDownlevelFlags, WaitIdleError},
+    global::Global,
+    hal_api::HalApi,
+    hal_label,
+    hub::Token,
     id::{DeviceId, SurfaceId, TextureId, Valid},
+    identity::{GlobalIdentityHandlerFactory, Input},
     init_tracker::TextureInitTracker,
-    resource,
-    track::TextureSelector,
-    LifeGuard, Stored,
+    resource, track, LifeGuard, Stored,
 };
 
 use hal::{Queue as _, Surface as _};
@@ -34,7 +36,7 @@ pub const DESIRED_NUM_FRAMES: u32 = 3;
 #[derive(Debug)]
 pub(crate) struct Presentation {
     pub(crate) device_id: Stored<DeviceId>,
-    pub(crate) config: wgt::SurfaceConfiguration,
+    pub(crate) config: wgt::SurfaceConfiguration<Vec<wgt::TextureFormat>>,
     #[allow(unused)]
     pub(crate) num_frames: u32,
     pub(crate) acquired_texture: Option<Stored<TextureId>>,
@@ -47,38 +49,66 @@ impl Presentation {
 }
 
 #[derive(Clone, Debug, Error)]
+#[non_exhaustive]
 pub enum SurfaceError {
-    #[error("surface is invalid")]
+    #[error("Surface is invalid")]
     Invalid,
-    #[error("surface is not configured for presentation")]
+    #[error("Surface is not configured for presentation")]
     NotConfigured,
     #[error(transparent)]
     Device(#[from] DeviceError),
-    #[error("surface image is already acquired")]
+    #[error("Surface image is already acquired")]
     AlreadyAcquired,
-    #[error("acquired frame is still referenced")]
+    #[error("Acquired frame is still referenced")]
     StillReferenced,
 }
 
 #[derive(Clone, Debug, Error)]
+#[non_exhaustive]
 pub enum ConfigureSurfaceError {
     #[error(transparent)]
     Device(#[from] DeviceError),
-    #[error("invalid surface")]
+    #[error("Invalid surface")]
     InvalidSurface,
+    #[error("The view format {0:?} is not compatible with texture format {1:?}, only changing srgb-ness is allowed.")]
+    InvalidViewFormat(wgt::TextureFormat, wgt::TextureFormat),
+    #[error(transparent)]
+    MissingDownlevelFlags(#[from] MissingDownlevelFlags),
     #[error("`SurfaceOutput` must be dropped before a new `Surface` is made")]
     PreviousOutputExists,
     #[error("Both `Surface` width and height must be non-zero. Wait to recreate the `Surface` until the window has non-zero area.")]
     ZeroArea,
-    #[error("surface does not support the adapter's queue family")]
+    #[error("Surface does not support the adapter's queue family")]
     UnsupportedQueueFamily,
-    #[error("requested format {requested:?} is not in list of supported formats: {available:?}")]
+    #[error("Requested format {requested:?} is not in list of supported formats: {available:?}")]
     UnsupportedFormat {
         requested: wgt::TextureFormat,
         available: Vec<wgt::TextureFormat>,
     },
-    #[error("requested usage is not supported")]
+    #[error("Requested present mode {requested:?} is not in the list of supported present modes: {available:?}")]
+    UnsupportedPresentMode {
+        requested: wgt::PresentMode,
+        available: Vec<wgt::PresentMode>,
+    },
+    #[error("Requested alpha mode {requested:?} is not in the list of supported alpha modes: {available:?}")]
+    UnsupportedAlphaMode {
+        requested: wgt::CompositeAlphaMode,
+        available: Vec<wgt::CompositeAlphaMode>,
+    },
+    #[error("Requested usage is not supported")]
     UnsupportedUsage,
+    #[error("Gpu got stuck :(")]
+    StuckGpu,
+}
+
+impl From<WaitIdleError> for ConfigureSurfaceError {
+    fn from(e: WaitIdleError) -> Self {
+        match e {
+            WaitIdleError::Device(d) => ConfigureSurfaceError::Device(d),
+            WaitIdleError::WrongSubmissionIndex(..) => unreachable!(),
+            WaitIdleError::StuckGpu => ConfigureSurfaceError::StuckGpu,
+        }
+    }
 }
 
 #[repr(C)]
@@ -94,7 +124,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         surface_id: SurfaceId,
         texture_id_in: Input<G, TextureId>,
     ) -> Result<SurfaceOutput, SurfaceError> {
-        profiling::scope!("get_next_texture", "SwapChain");
+        profiling::scope!("SwapChain::get_next_texture");
 
         let hub = A::hub(self);
         let mut token = Token::root();
@@ -109,6 +139,9 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let (device, config) = match surface.presentation {
             Some(ref present) => {
                 let device = &device_guard[present.device_id.value];
+                if !device.is_valid() {
+                    return Err(DeviceError::Lost.into());
+                }
                 (device, present.config.clone())
             }
             None => return Err(SurfaceError::NotConfigured),
@@ -125,10 +158,19 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let _ = device;
 
         let suf = A::get_surface_mut(surface);
-        let (texture_id, status) = match unsafe { suf.raw.acquire_texture(FRAME_TIMEOUT_MS) } {
+        let (texture_id, status) = match unsafe {
+            suf.unwrap()
+                .raw
+                .acquire_texture(Some(std::time::Duration::from_millis(
+                    FRAME_TIMEOUT_MS as u64,
+                )))
+        } {
             Ok(Some(ast)) => {
                 let clear_view_desc = hal::TextureViewDescriptor {
-                    label: Some("clear surface texture view"),
+                    label: hal_label(
+                        Some("(wgpu internal) clear surface texture view"),
+                        self.instance.flags,
+                    ),
                     format: config.format,
                     dimension: wgt::TextureViewDimension::D2,
                     usage: hal::TextureUses::COLOR_TARGET,
@@ -166,17 +208,18 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                         format: config.format,
                         dimension: wgt::TextureDimension::D2,
                         usage: config.usage,
+                        view_formats: config.view_formats,
                     },
                     hal_usage: conv::map_texture_usage(config.usage, config.format.into()),
                     format_features: wgt::TextureFormatFeatures {
                         allowed_usages: wgt::TextureUsages::RENDER_ATTACHMENT,
-                        flags: wgt::TextureFormatFeatureFlags::empty(),
-                        filterable: false,
+                        flags: wgt::TextureFormatFeatureFlags::MULTISAMPLE_X4
+                            | wgt::TextureFormatFeatureFlags::MULTISAMPLE_RESOLVE,
                     },
                     initialization_status: TextureInitTracker::new(1, 1),
-                    full_range: TextureSelector {
+                    full_range: track::TextureSelector {
                         layers: 0..1,
-                        levels: 0..1,
+                        mips: 0..1,
                     },
                     life_guard: LifeGuard::new("<Surface>"),
                     clear_mode: resource::TextureClearMode::RenderPass {
@@ -188,7 +231,15 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                 let ref_count = texture.life_guard.add_ref();
                 let id = fid.assign(texture, &mut token);
 
-                //suf.acquired_texture = Some(suf_texture);
+                {
+                    // register it in the device tracker as uninitialized
+                    let mut trackers = device.trackers.lock();
+                    trackers.textures.insert_single(
+                        id.0,
+                        ref_count.clone(),
+                        hal::TextureUses::UNINITIALIZED,
+                    );
+                }
 
                 if present.acquired_texture.is_some() {
                     return Err(SurfaceError::AlreadyAcquired);
@@ -229,7 +280,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         &self,
         surface_id: SurfaceId,
     ) -> Result<Status, SurfaceError> {
-        profiling::scope!("present", "SwapChain");
+        profiling::scope!("SwapChain::present");
 
         let hub = A::hub(self);
         let mut token = Token::root();
@@ -246,6 +297,9 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         };
 
         let device = &mut device_guard[present.device_id.value];
+        if !device.is_valid() {
+            return Err(DeviceError::Lost.into());
+        }
 
         #[cfg(feature = "trace")]
         if let Some(ref trace) = device.trace {
@@ -260,19 +314,15 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
 
             // The texture ID got added to the device tracker by `submit()`,
             // and now we are moving it away.
+            log::debug!(
+                "Removing swapchain texture {:?} from the device tracker",
+                texture_id.value
+            );
             device.trackers.lock().textures.remove(texture_id.value);
 
             let (texture, _) = hub.textures.unregister(texture_id.value.0, &mut token);
             if let Some(texture) = texture {
-                if let resource::TextureClearMode::RenderPass { clear_views, .. } =
-                    texture.clear_mode
-                {
-                    for clear_view in clear_views {
-                        unsafe {
-                            hal::Device::destroy_texture_view(&device.raw, clear_view);
-                        }
-                    }
-                }
+                texture.clear_mode.destroy_clear_views(&device.raw);
 
                 let suf = A::get_surface_mut(surface);
                 match texture.inner {
@@ -286,10 +336,10 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                             Err(hal::SurfaceError::Lost)
                         } else if !has_work {
                             log::error!("No work has been submitted for this frame");
-                            unsafe { suf.raw.discard_texture(raw) };
+                            unsafe { suf.unwrap().raw.discard_texture(raw) };
                             Err(hal::SurfaceError::Outdated)
                         } else {
-                            unsafe { device.queue.present(&mut suf.raw, raw) }
+                            unsafe { device.queue.present(&mut suf.unwrap().raw, raw) }
                         }
                     }
                     resource::TextureInner::Native { .. } => unreachable!(),
@@ -319,7 +369,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         &self,
         surface_id: SurfaceId,
     ) -> Result<(), SurfaceError> {
-        profiling::scope!("discard", "SwapChain");
+        profiling::scope!("SwapChain::discard");
 
         let hub = A::hub(self);
         let mut token = Token::root();
@@ -336,6 +386,9 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         };
 
         let device = &mut device_guard[present.device_id.value];
+        if !device.is_valid() {
+            return Err(DeviceError::Lost.into());
+        }
 
         #[cfg(feature = "trace")]
         if let Some(ref trace) = device.trace {
@@ -350,10 +403,16 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
 
             // The texture ID got added to the device tracker by `submit()`,
             // and now we are moving it away.
+            log::debug!(
+                "Removing swapchain texture {:?} from the device tracker",
+                texture_id.value
+            );
             device.trackers.lock().textures.remove(texture_id.value);
 
             let (texture, _) = hub.textures.unregister(texture_id.value.0, &mut token);
             if let Some(texture) = texture {
+                texture.clear_mode.destroy_clear_views(&device.raw);
+
                 let suf = A::get_surface_mut(surface);
                 match texture.inner {
                     resource::TextureInner::Surface {
@@ -362,7 +421,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                         has_work: _,
                     } => {
                         if surface_id == parent_id.0 {
-                            unsafe { suf.raw.discard_texture(raw) };
+                            unsafe { suf.unwrap().raw.discard_texture(raw) };
                         } else {
                             log::warn!("Surface texture is outdated");
                         }
